! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module param

  use pil_f90_api
  use iso_fortran_env

  implicit none

  integer, parameter, public :: niter = 19 ! max. range of loop indices

  public :: get_sass_history, &
    opast, cpast, wrlog, wrchatty, wrerr, wrstop, fitserr, rputc, &
    rputi, sputi, sputc,  rget, rgetc, rgetj, rgeti, rgetf, rgetd, &
    rgetl, rgeta, rgetca, rgetia, rgetja, rgetfa, rgetda, rgetla, &
    sgetc, pgeti, pgetc, rgetn, lcase, errstc, prep_gtcalf, lock, unlock

  interface rget
    module procedure rgetc2, rgetd2, rgetf2, rgeti2, rgetj2, rgetl2
  end interface

  interface rgeta
    module procedure rgetca2, rgetda2, rgetfa2, rgetia2, rgetja2, rgetla2
  end interface

  interface opast
    module procedure opast1, opast2
  end interface

  interface cpast
    module procedure cpast1, cpast2
  end interface

  interface wrstop
    module procedure wrstop1, wrstop2
  end interface

  interface rputc
    module procedure rputc1, rputc2
  end interface

  interface rputi
    module procedure rputi1, rputi2
  end interface

  interface sputc
    module procedure sputc1, sputc2
  end interface

  interface sputi
    module procedure sputi1, sputi2
  end interface

  interface rgetc
    module procedure rgetc1, rgetc2
  end interface

  interface rgeti
    module procedure rgeti1, rgeti2
  end interface

  interface rgetj
    module procedure rgetj1, rgetj2
  end interface

  interface rgetf
    module procedure rgetf1, rgetf2
  end interface

  interface rgetd
    module procedure rgetd1, rgetd2
  end interface

  interface rgetl
    module procedure rgetl1, rgetl2
  end interface

  interface rgetca
    module procedure rgetca1, rgetca2
  end interface

  interface rgetia
    module procedure rgetia1, rgetia2
  end interface

  interface rgetja
    module procedure rgetja1, rgetja2
  end interface

  interface rgetfa
    module procedure rgetfa1, rgetfa2
  end interface

  interface rgetda
    module procedure rgetda1, rgetda2
  end interface

  interface rgetla
    module procedure rgetla1, rgetla2
  end interface

  interface sgetc
    module procedure sgetc1, sgetc2
  end interface

  interface prep_gtcalf
    module procedure prep_gtcalf1, prep_gtcalf2
  end interface

  ! The PIL library is missing F90 interface to PILVerifyCmdLine,
  ! to check the command line options.
  ! This is a wrapper for the C function to use it here
  interface
    function C_PILVerifyCmdLine() bind(C, name="PILVerifyCmdLine")
      use iso_c_binding, only: C_INT
      integer(kind=C_INT) :: C_PILVerifyCmdLine
    end function C_PILVerifyCmdLine
  end interface

  integer, parameter :: logpar=14
  logical :: pil = .false., rgetn_first = .true.
  character(len=16384), public :: sass_history = " "

contains

  !-------------------------------------------------------------
  !   function get_sass_history() result (sass_hist_for_c)
  !-------------------------------------------------------------
  !
  !     character(len=16384) :: sass_hist_for_c
  !     sass_hist_for_c = sass_history
  !
  !   end function get_sass_history
  !-------------------------------------------------------------

  !-------------------------------------------------------------
  subroutine get_sass_history (sass_hist_for_c) bind (C, name="get_sass_history")
    !-------------------------------------------------------------

    use iso_c_binding, only: C_CHAR, C_NULL_CHAR

    character(len=1, kind=C_CHAR), dimension (16384), intent (out) :: sass_hist_for_c

    character(len=16384, kind=C_CHAR) :: string
    integer :: i

    string = sass_history

    do i=1, 16384
      sass_hist_for_c (i) = string (i:i)
    end do

    return

  end subroutine get_sass_history

  !-------------------------------------------------------------
  subroutine wait_open(unit)
    !-------------------------------------------------------------

    integer, intent(in) :: unit
    logical :: isopen = .true.

    integer, parameter :: waitsec = 1

    do while (isopen)
      call sleep (waitsec)
      inquire(unit, opened=isopen)
    end do
  
  end subroutine wait_open

  !-------------------------------------------------------------
  subroutine lock(dir, lockfile)
    !-------------------------------------------------------------

    character(*), intent(in) :: dir, lockfile
    integer, parameter :: waitsec=1, maxlock=30
    integer :: i

    character(len=256) :: file
    character(len=3) :: c3
    logical :: ex

    ! CG
    integer :: unit
    integer :: status
    integer :: pid
    integer :: pid_lock
    integer :: pid_lock_save = 0
    integer :: locked = 0
    integer :: counter = 0
    character(6) :: c_pid
    character(255) :: cwd

    file = trim(dir) // "/" // trim(lockfile)
    pid = getpid()
    write(c_pid, '(I6.6)') pid

    !CG directly open the file instead of query and touch later
    do while (locked == 0)
      ! forcefully unlock
      if(counter > maxlock) then
        call wrerr("lock", 1, "Lock file " // trim(file) // " forcefully removed")
        call unlock(dir, lockfile)
        ! reset counter in case other tasks are faster
        counter = 0
      end if

      open(newunit=unit, file=file, status = "new", action="write", iostat=status)
      ! if lock was created write pid of the process which created it to the LOCK
      if(status == 0) then
        ! write pid
        write(unit, *, iostat=status) pid
        close(unit, iostat=status)

        ! only if nothing failed until now can someone leave the lock
        if(status == 0) then
          if(dir == ".") then
            call getcwd(cwd)
            file = trim(cwd) // "/" // trim(lockfile)
          end if
!          call wrlog("LOCK", trim(file) // " with PID " // trim(c_pid))
          return
        end if
      end if

      open(newunit=unit, file=file, status = "old", action="read", iostat=status)
      if(status == 0) then
        read(unit, *, iostat=status) pid_lock
        if(status == 0) then
          if(pid_lock /= pid_lock_save) then
            counter = 0
            pid_lock_save = pid_lock
          end if
        end if
      end if

      call sleep(waitsec)
      counter = counter + 1
    end do

  end subroutine lock

  !-------------------------------------------------------------
  subroutine unlock(dir, lockfile)
    !-------------------------------------------------------------

    character(len=*), intent(in) :: dir, lockfile
    character(len=256) :: file
    integer :: pid
    character(6) :: c_pid
    character(255) :: cwd

    file = trim(dir) // "/" // trim(lockfile)
    call unlink(file)

    pid = getpid()
    write(c_pid, '(I6.6)') pid
    if(dir == ".") then
      call getcwd(cwd)
      file = trim(cwd) // "/" // trim(lockfile)
    end if
!    call wrlog("UNLOCK", trim(file) // " with PID " // trim(c_pid))

  end subroutine unlock

  !-------------------------------------------------------------
  subroutine opast1(rtname, lupar, ierr)
    !-------------------------------------------------------------

    character(len=*), intent(in) :: rtname
    integer, intent(in) :: lupar
    integer, intent(out) :: ierr

    integer :: narg, ind1, ind2, len, istat
    integer(kind=8) :: it8
    character(len=1) :: cind
    character(len=8) :: par_int
    character(len=32)  :: task, date

    character(len=64) :: version

    ierr = 0

    call wrlog("",trim(version()))

    ! PIL or pipeline parameter interface?
    call get_environment_variable('SASS_PARINT', par_int)
    if (par_int .ne. 'PIPELINE' .and. par_int .ne. 'pipeline' ) then
      istat = PILinit()
      if (ISTAT == 0) then
        pil = .true.
        !      call wrlog("","STARTED")
        call wrchatty ("opast", &
          "**** Using command line parameter interface ****" )
      else
        ierr=4
        call wrerr("opast",ierr,'Error initializing PIL')
      end if

      ! check the parameters match what are in the .par file
      istat = C_PILVerifyCmdLine()
      if (istat /= 0) then
         ierr=5
         call wrerr("opast",ierr,'PIL command line verification failed')
      end if

    else
      call wrchatty ("opast", & 
        "**** Using pipeline parameter interface ****")
    end if
      
  end subroutine opast1

  !-------------------------------------------------------------
  subroutine opast2(ierr)
    !-------------------------------------------------------------

    integer, intent(out) :: ierr

    integer :: narg, ind1, ind2, len, istat
    integer(kind=8) :: it8
    character(len=1) :: cind
    character(len=8) :: par_int
    character(len=32)  :: task, date

    character(len=64) :: version

    ierr = 0

    call wrlog("",trim(version()))

    ! PIL or pipeline parameter interface?
    call get_environment_variable('SASS_PARINT', par_int)
    if (par_int .ne. 'PIPELINE' .and. par_int .ne. 'pipeline' ) then
      istat = PILinit()
      if (ISTAT == 0) then
        pil = .true.
        !      call wrlog("","STARTED")
        call wrchatty ("opast", &
          "**** Using command line parameter interface ****" )
      else
        ierr=4
        call wrerr("opast",ierr,'Error initializing PIL')
      end if

      ! check the parameters match what are in the .par file
      istat = C_PILVerifyCmdLine()
      if (istat /= 0) then
         ierr=5
         call wrerr("opast",ierr,'PIL command line verification failed')
      end if

    else
      call wrchatty ("opast", & 
        "**** Using pipeline parameter interface ****")
    end if
      
  end subroutine opast2

  !-------------------------------------------------------------
  subroutine cpast1(rtname, lupar, status, ierr)
    !-------------------------------------------------------------
    ! sets STAT paramter to START in SASS.PAR
    ! (also done in program exec, so perhaps not needed here)

    character(len=*), intent(in) :: rtname
    integer, intent(in) :: lupar, status
    integer, intent(out) :: ierr

    integer :: ind1, ind2, ind3, ind4, len, istat, iostat
    integer(kind=8) :: it8
    character(len=2) :: cind
    character(len=8) :: par_int
    character(len=32)  :: task
    character(len=40) :: date

    ierr = 0

    call get_environment_variable('SASS_PARINT', par_int)
    if (par_int .eq. 'PIPELINE' .or. par_int .eq. 'pipeline' ) then
      !  if (.not. pil) then

      ! get loop indices from command line
      call get_command_argument(1, cind, len, istat)
      if (istat==0 .and. len==1) then
        read(cind,'(i1)',iostat=iostat) ind1
      else if (istat==0 .and. len==2) then
        read(cind,'(i2)',iostat=iostat) ind1
      else
        ind1=0
      end if
      if (iostat/=0) ind1=0

      call get_command_argument(2, cind, len, istat)
      if (istat==0 .and. len==1) then
        read(cind,'(i1)',iostat=iostat) ind2
      else if (istat==0 .and. len==2) then
        read(cind,'(i2)',iostat=iostat) ind2
      else
        ind2=0
      end if
      if (iostat/=0) ind2=0

      call get_command_argument(3, cind, len, istat)
      if (istat==0 .and. len==1) then
        read(cind,'(i1)',iostat=iostat) ind3
      else if (istat==0 .and. len==2) then
        read(cind,'(i2)',iostat=iostat) ind3
      else
        ind3=0
      end if
      if (iostat/=0) ind3=0

      call get_command_argument(4, cind, len, istat)
      if (istat==0 .and. len==1) then
        read(cind,'(i1)',iostat=iostat) ind4
      else if (istat==0 .and. len==2) then
        read(cind,'(i2)',iostat=iostat) ind4
      else
        ind4=0
      end if
      if (iostat/=0) ind4=0

      call get_command_argument(0, task)

      it8=time8()
      call ctime(it8, date)
      if (status>=0 .and. status < 3) then
        call wrlog("","DONE")
        call sputc("STAT", trim(task), ind1, ind2, ind3, ind4, "DONE: " // trim(date(5:)), ierr)
        ! CG return - no need to check for status again
        return
      else
        call wrlog("","FAILED")
        call sputc("STAT", trim(task), ind1, ind2, ind3, ind4, "ERROR: " // trim(date(5:)), ierr)
        call exit(1)
      end if

    end if

    ! CG in non-pipeline mode check for status
    if(status >= 0 .and. status < 3) then
      call wrlog("","DONE")
    else
      call wrlog("","FAILED")
      call exit(1)
    end if

  end subroutine cpast1

  !-------------------------------------------------------------
  subroutine cpast2(status, ierr)
    !-------------------------------------------------------------
    ! sets STAT paramter to START in SASS.PAR
    ! (also done in program exec, so perhaps not needed here)

    integer, intent(in) :: status
    integer, intent(out) :: ierr

    integer :: ind1, ind2, ind3, ind4, len, istat, iostat
    integer(kind=8) :: it8
    character(len=2) :: cind
    character(len=8) :: par_int
    character(len=32)  :: task
    character(len=40) :: date

    ierr = 0

    call get_environment_variable('SASS_PARINT', par_int)
    if (par_int .eq. 'PIPELINE' .or. par_int .eq. 'pipeline' ) then
      !  if (.not. pil) then

      ! get loop indices from command line
      call get_command_argument(1, cind, len, istat)
      if (istat==0 .and. len==1) then
        read(cind,'(i1)',iostat=iostat) ind1
      else if (istat==0 .and. len==2) then
        read(cind,'(i2)',iostat=iostat) ind1
      else
        ind1=0
      end if
      if (iostat/=0) ind1=0

      call get_command_argument(2, cind, len, istat)
      if (istat==0 .and. len==1) then
        read(cind,'(i1)',iostat=iostat) ind2
      else if (istat==0 .and. len==2) then
        read(cind,'(i2)',iostat=iostat) ind2
      else
        ind2=0
      end if
      if (iostat/=0) ind2=0

      call get_command_argument(3, cind, len, istat)
      if (istat==0 .and. len==1) then
        read(cind,'(i1)',iostat=iostat) ind3
      else if (istat==0 .and. len==2) then
        read(cind,'(i2)',iostat=iostat) ind3
      else
        ind3=0
      end if
      if (iostat/=0) ind3=0

      call get_command_argument(4, cind, len, istat)
      if (istat==0 .and. len==1) then
        read(cind,'(i1)',iostat=iostat) ind4
      else if (istat==0 .and. len==2) then
        read(cind,'(i2)',iostat=iostat) ind4
      else
        ind4=0
      end if
      if (iostat/=0) ind4=0

      call get_command_argument(0, task)

      it8=time8()
      call ctime(it8, date)
      if (status >=0 .and. status < 3) then
        call wrlog("","DONE")
        call sputc("STAT", task, ind1, ind2, ind3, ind4, "DONE:  " // trim(date(5:)), ierr)
        ! CG return no need to check again
        return
      else
        call wrlog("","FAILED")
        call sputc("STAT", task, ind1, ind2, ind3, ind4, "ERROR: " // trim(date(5:)), ierr)
        call exit(1)
      end if

    end if

    ! CG in non-pipeline mode check for status
    if(status>=0 .and. status < 3) then
      call wrlog("","DONE")
    else
      call wrlog("","FAILED")
      call exit(1)
    end if

  end subroutine cpast2

  !---------------------------------------------------------------
  subroutine wrlog(rtname, text)
    !---------------------------------------------------------------

    character(len=*), intent(in) :: rtname, text
    integer(kind=8) :: it8
    integer :: status, prpos
    character(len=8) :: verbosity, par_int, date8
    character(len=32) :: date, chost, cname, file, rtnam
    character(len=128) :: dir, eroarch

    call get_environment_variable('SASS_VERBOSITY', verbosity)
    if (verbosity /= "SPARSE" .and. verbosity /= "sparse") then
      it8=time8()
      call ctime(it8, date)
      ! remove week day and year from date string
      date = date(5:19)
      call hostnm (chost, status)
      call getcwd(dir)
      prpos=index(dir,"products")
      if (prpos>0) then
        dir=trim(dir(prpos+9:)) // ":"
      else
        dir=""
      end if
      call get_command_argument(0, cname)
      if (len_trim(rtname) > 0) then
        rtnam = "/" // rtname
      else
        rtnam=""
      end if
      call get_environment_variable('SASS_PARINT', par_int)
      if (par_int == 'PIPELINE' .or. par_int == 'pipeline' ) then
        call get_environment_variable('ERO_ARCHIVE', eroarch)
        if (trim(cname) /= "eropipe" .and. trim(cname) /= "prexec" .and. trim(cname) /= "proc" .or. &
          len_trim(eroarch)==0) then
          dir="."
          file="/SASS.LOG"
        else
          call date_and_time(date8)
          dir=trim(eroarch) // "/config/log"
          file ="/" // date8 // ".log"
        end if
        !!!call lock(dir,"LOCK")
        call wait_open(logpar)
        open(unit=logpar, file=trim(dir)//trim(file), form="formatted", &
          status="unknown", position="append")
        write(logpar,'(a)') trim(date) // " " // trim(chost) // ":" // trim(dir) // trim(cname) // trim(rtnam) // ": " // &
          trim(text)
        close(logpar)
        !!!call unlock(dir,"LOCK")
      else
        print *, trim(cname) // trim(rtnam) // ": " // trim(text)
      end if
    end if

  end subroutine wrlog

  !---------------------------------------------------------------
  subroutine wrchatty(rtname, text)
    !---------------------------------------------------------------

    character(len=*), intent(in) :: rtname, text
    integer(kind=8) :: it8
    integer :: status, prpos
    character(len=8) :: verbosity, par_int, date8
    character(len=32) :: date, chost, cname, file, rtnam
    character(len=128) :: dir

    call get_environment_variable('SASS_VERBOSITY', verbosity)
    if (verbosity == "CHATTY" .or. verbosity == "chatty") then
      it8=time8()
      call ctime(it8, date)
      ! remove week day and year from date string
      date = date(5:19)
      call hostnm (chost, status)
      call getcwd(dir)
      prpos=index(dir,"products")
      if (prpos>0) then
        dir=trim(dir(prpos+9:)) // ":"
      else
        dir=""
      end if
      call get_command_argument(0, cname)
      if (len_trim(rtname) > 0) then
        rtnam = "/" // trim(rtname)
      else
        rtnam=""
      end if
      call get_environment_variable('SASS_PARINT', par_int)
      if (par_int == 'PIPELINE' .or. par_int == 'pipeline' ) then
        if (trim(cname) /= "eropipe" .and. trim(cname) /= "prexec" .and. trim(cname) /= "proc") then
          dir="."
          file="/SASS.LOG"
        else
          call date_and_time(date8)
          dir="$ERO_ARCHVIVE/config/log"
          file="/" // date8 // ".log"
        end if
        !!!call lock(dir,"LOCK")
        call wait_open(logpar)
        open(unit=logpar, file=trim(dir)//trim(file), form="formatted", &
          status="unknown", position="append")
        write(logpar,'(a)') trim(date) // " " // trim(chost) // ":" // trim(dir) // trim(cname) // trim(rtnam) // ": " // &
          trim(text)
        close(logpar)
        !!!call unlock(dir,"LOCK")
      else
        print *, trim(cname) // trim(rtnam) // ": " // trim(text)
      end if
    end if

  end subroutine wrchatty

  !---------------------------------------------------------------
  subroutine wrerr(rtname, ierr, text)
    !---------------------------------------------------------------

    character(len=*), intent(in) :: rtname, text
    integer, intent(in) :: ierr
    integer(kind=8) :: it8
    integer :: status, prpos
    character(len=8) :: verbosity, par_int, date8
    character(len=13) :: cerr
    character(len=32) :: date, chost, cname, file, rtnam
    character(len=128) :: dir, eroarch

    ! new error level 4: fatal error, will stop immediately
    ! (level 3 is also a fatal error, but wrerr will return to the
    ! calling routine to allow the program to terminate gracefully)
    if (ierr>=4) call wrstop(999, rtname, text)

    it8=time8()
    call ctime(it8, date)
    ! remove week day and year from date string
    date = date(5:19)
    call hostnm (chost, status)
    call getcwd(dir)
    prpos=index(dir,"products")
    if (prpos>0) then
      dir=trim(dir(prpos+9:)) // ":"
    else
      dir=""
    end if 
    call get_command_argument(0, cname)
    write(cerr,'(a,i1,a)') "**ERROR", ierr, "** "
    ! ERROR0/1 is warning, so write WARNING instead
    if (ierr<=1) cerr="**WARNING" // cerr(8:8) // "** "
    if (len_trim(rtname) > 0) then
      rtnam = "/" // trim(rtname)
    else
      rtnam=""
    end if
    call get_environment_variable('SASS_PARINT', par_int)
    if (par_int == 'PIPELINE' .or. par_int == 'pipeline' ) then
      call get_environment_variable('ERO_ARCHIVE', eroarch)
      if (trim(cname) /= "eropipe" .and. trim(cname) /= "prexec" .and. trim(cname) /= "proc" .or. &
        len_trim(eroarch)==0) then
        dir="."
        file="/SASS.LOG"
      else
        call date_and_time(date8)
        dir = trim(eroarch) // "/config/log"
        file = "/" // date8 // ".log"
      end if
      !!!call lock(dir,"LOCK")
      call wait_open(logpar)
      open(unit=logpar, file=trim(dir)//trim(file), form="formatted", &
        status="unknown", position="append")
      write(logpar,'(a)') trim(date) // " " // trim(chost) // ":" // trim(dir) // trim(cname) // trim(rtnam) // ": " // &
        trim(cerr) // " " // trim(text)
      close(logpar)
      !!!call unlock(dir,"LOCK")
    else
      print *, trim(cname) // trim(rtnam) // ": " // trim(cerr) // " " // trim(text)
    end if

  end subroutine wrerr

  !---------------------------------------------------------------
  subroutine wrstop1(lupar,rtname,text)
    !---------------------------------------------------------------

    character(len=*), intent(in) :: rtname, text
    integer, intent(in) :: lupar
    integer :: ierr, status, prpos
    integer(kind=8) :: it8
    character(len=8) :: verbosity, par_int, date8
    character(len=10) :: cerr
    character(len=32) :: date, chost, cname, file, rtnam
    character(len=128) :: dir, eroarch

    it8=time8()
    call ctime(it8, date)
    ! remove week day and year from date string
    date = date(5:19)   
    call hostnm (chost, status)
    call getcwd(dir)
    prpos=index(dir,"products")
    if (prpos>0) then
      dir=trim(dir(prpos+9:)) // ":"
    else
      dir=""
    end if 
    call get_command_argument(0, cname)
    if (len_trim(rtname) > 0) then
      rtnam = "/" // trim(rtname)
    else
      rtnam=""
    end if
    call get_environment_variable('SASS_PARINT', par_int)
    if (par_int == 'PIPELINE' .or. par_int == 'pipeline' ) then
      call get_environment_variable('ERO_ARCHIVE', eroarch)
      if (trim(cname) /= "eropipe" .and. trim(cname) /= "prexec" .and. trim(cname) /= "proc" .or. &
        len_trim(eroarch)==0) then
        dir="."
        file="/SASS.LOG"
      else
        call date_and_time(date8)
        dir= trim(eroarch) // "/config/log"
        file= "/" // date8 // ".log"
      end if
      !!!call lock(dir,"LOCK")
      call wait_open(logpar)
      open(unit=logpar, file=trim(dir)//trim(file), form="formatted", &
        status="unknown", position="append")
      write(logpar,'(a)') trim(date) // " " // trim(chost) // ":" // trim(dir) // trim(cname) // trim(rtnam) // ": " // &
        "**STOP** " // trim(text)
      close(logpar)
      !!!call unlock(dir,"LOCK")
    else
      print *, trim(cname) // trim(rtnam) // ": " // "**STOP** " // trim(text)
    end if

    call cpast(rtname, lupar, 3, ierr)
    call exit(1)

  end subroutine wrstop1

  !---------------------------------------------------------------
  subroutine wrstop2(rtname,text)
    !---------------------------------------------------------------

    character(len=*), intent(in) :: rtname, text
    integer :: unit, ierr, status, prpos
    integer(kind=8) :: it8
    character(len=8) :: verbosity, par_int, date8
    character(len=10) :: cerr
    character(len=32) :: date, chost, cname, file, rtnam
    character(len=128) :: dir, eroarch

    it8=time8()
    call ctime(it8, date)
    ! remove week day and year from date string
    date = date(5:19)
    call hostnm (chost, status)
    call getcwd(dir)
    prpos=index(dir,"products")
    if (prpos>0) then
      dir=trim(dir(prpos+9:)) // ":"
    else
      dir=""
    end if
    call get_command_argument(0, cname)
    if (len_trim(rtname) > 0) then
      rtnam = "/" // trim(rtname)
    else
      rtnam=""
    end if
    call get_environment_variable('SASS_PARINT', par_int)
    if (par_int == 'PIPELINE' .or. par_int == 'pipeline' ) then
      call get_environment_variable('ERO_ARCHIVE', eroarch)
      if (trim(cname) /= "eropipe" .and. trim(cname) /= "prexec" .and. trim(cname) /= "proc" .or. &
        len_trim(eroarch)==0) then
        dir="."
        file="/SASS.LOG"
      else
        call date_and_time(date8)
        dir= trim(eroarch) // "/config/log"
        file= "/" // date8 // ".log"
      end if
      !!!call lock(dir,"LOCK")
      call wait_open(logpar)
      open(newunit=unit, file=trim(dir)//trim(file), form="formatted", &
        status="unknown", position="append")
      write(unit,'(a)') trim(date) // " " // trim(chost) // ":" // trim(dir) // trim(cname) // trim(rtnam) // ": " // &
        "**STOP** " // trim(text)
      close(unit)
      !!!call unlock(dir,"LOCK")
    else
      print *, trim(cname) // trim(rtnam) // ": " // "**STOP** " // trim(text)
    end if

    call cpast(3, ierr)
    call exit(1)

  end subroutine wrstop2

  !---------------------------------------------------------------
  subroutine fitserr(rtname, status, ierr, comment)
    !---------------------------------------------------------------

    character(len=*), intent(in) :: rtname
    integer, intent(inout) :: status
    integer, intent(out) :: ierr
    character(len=*), intent(in), optional :: comment

    character(len=30) :: errtext

    ierr = 0 !*** but it can't be anything else. Could get rid of this argument.

    if (status>0) then
      call ftgerr(status, errtext)
      if(present(comment)) then
        call wrstop(rtname, trim(errtext)//trim(comment))
      else
        call wrstop(rtname, trim(errtext))
      end if
      status = 0
    else
      ierr=0
    end if

  end subroutine fitserr

  !---------------------------------------------------------------
  subroutine rputc1(lupar, param, c, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar
    integer, intent(out) :: ierr
    character(len=*), intent(in) :: param, c

    ierr = 0 !*** but it can't be anything else. Could get rid of this argument.

    call lock(".","LOCK.PAR")
    open(unit=lupar, file="SASS.PAR", form="formatted", status="old", position="append")
    write(lupar,'(3a)') trim(param), "=", trim(c)
    close(lupar)
    call unlock(".","LOCK.PAR")

  end subroutine rputc1

  !---------------------------------------------------------------
  subroutine rputc2(param, c, ierr)
    !---------------------------------------------------------------

    integer, intent(out) :: ierr
    character(len=*), intent(in) :: param, c
    integer :: unit

    ierr = 0 !*** but it can't be anything else. Could get rid of this argument.

    call lock(".","LOCK.PAR")
    open(newunit=unit, file="SASS.PAR", form="formatted", status="old", position="append")
    write(unit,'(3a)') trim(param), "=", trim(c)
    close(unit)
    call unlock(".","LOCK.PAR")

  end subroutine rputc2

  !---------------------------------------------------------------
  subroutine rputi1(lupar, param, ival, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar, ival
    integer, intent(out) :: ierr
    character(len=*), intent(in) :: param
    character(len=11) :: cval

    ierr = 0 !*** but it can't be anything else. Could get rid of this argument.

    call lock(".","LOCK.PAR")
    open(unit=lupar, file="SASS.PAR", form="formatted", status="old", position="append")
    write(cval,'(i11)') ival
    write(lupar,'(3a)') trim(param), "=", trim(adjustl(cval))
    close(lupar)
    call unlock(".","LOCK.PAR")

  end subroutine rputi1

  !---------------------------------------------------------------
  subroutine rputi2(param, ival, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: ival
    integer, intent(out) :: ierr
    character(len=*), intent(in) :: param
    character(len=11) :: cval
    integer :: unit

    ierr = 0 !*** but it can't be anything else. Could get rid of this argument.

    call lock(".","LOCK.PAR")
    open(newunit=unit, file="SASS.PAR", form="formatted", status="old", position="append")
    write(cval,'(i11)') ival
    write(unit,'(3a)') trim(param), "=", trim(adjustl(cval))
    close(unit)
    call unlock(".","LOCK.PAR")

  end subroutine rputi2

  !---------------------------------------------------------------
  subroutine sputi1(lupar, param, task, ind1, ind2, ival, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar, ind1, ind2, ival
    integer, intent(out) :: ierr
    character(len=*), intent(in) :: param, task
    character(len=64) :: parnam
    character(len=11) :: cval
    character(len=1) :: ci

    ierr = 0 !*** but it can't be anything else. Could get rid of this argument.

    parnam = trim(adjustl(param))
    if (len(trim(task))>0) then
      parnam = parnam // "@" // trim(task)
      if (ind1>0 ) then
        write(ci, '(i1)') ind1
        parnam = parnam // "[" // CI // "]"
        if (ind2>0) then
          write(ci, '(i1)') ind1
          parnam = parnam // "[" // CI // "]"
        end if
      end if
    end if

    call lock(".","LOCK.PAR")
    open(unit=lupar, file="SASS.PAR", form="formatted", status="old", position="append")
    write(cval,'(i11)') ival
    write(lupar,'(3a)') parnam, "=", trim(adjustl(cval))
    close(lupar)
    call unlock(".","LOCK.PAR")

  end subroutine sputi1

  !---------------------------------------------------------------
  subroutine sputi2(param, task, ind1, ind2, ind3, ind4, ival, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: ind1, ind2, ind3, ind4, ival
    integer, intent(out) :: ierr
    character(len=*), intent(in) :: param, task
    character(len=64) :: parnam
    character(len=11) :: cval
    character(len=2) :: ci
    integer :: unit

    ierr = 0 !*** but it can't be anything else. Could get rid of this argument.

    if (ind1>niter .or. ind2>niter .or. ind3>niter .or. ind4>niter) call wrerr("sputi2", 4, "Index out of range")

    parnam = trim(adjustl(param))
    if (len(trim(task))>0) then
      parnam = trim(parnam) // "@" // trim(task)
      if (ind1>0 ) then
        if (ind1<10) then
          write(ci, '(i1)') ind1
          parnam = trim(parnam) // "[" // ci(1:1) // "]"
        else
          write(ci, '(i2)') ind1
          parnam = trim(parnam) // "[" // ci // "]"
        end if
        if (ind2>0) then
          if (ind2<10) then
            write(ci, '(i1)') ind2
            parnam = trim(parnam) // "[" // ci(1:1) // "]"
          else
            write(ci, '(i2)') ind2
            parnam = trim(parnam) // "[" // ci // "]"
          end if
          if (ind3>0) then
            if (ind3<10) then
              write(ci, '(i1)') ind3
              parnam = trim(parnam) // "[" // ci(1:1) // "]"
            else
              write(ci, '(i2)') ind3
              parnam = trim(parnam) // "[" // ci // "]"
            end if
            if (ind4>0) then
              if (ind4<10) then
                write(ci, '(i1)') ind4
                parnam = trim(parnam) // "[" // ci(1:1) // "]"
              else
                write(ci, '(i2)') ind4
                parnam = trim(parnam) // "[" // ci // "]"
              end if
            end if
          end if
        end if
      end if
    end if

    call lock(".","LOCK.PAR")
    open(newunit=unit, file="SASS.PAR", form="formatted", status="old", position="append")
    write(cval,'(i11)') ival
    write(unit,'(3a)') parnam, "=", trim(adjustl(cval))
    close(unit)
    call unlock(".","LOCK.PAR")

  end subroutine sputi2

  !---------------------------------------------------------------
  subroutine sputc1(lupar, param, task, ind1, ind2, c, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar, ind1, ind2
    integer, intent(out) :: ierr
    character(len=*), intent(in) :: param, task, c
    character(len=64) :: parnam
    character(len=1) :: ci

    ierr = 0 !*** but it can't be anything else. Could get rid of this argument.

    parnam = trim(adjustl(param))
    if (len(trim(task))>0) then
      parnam = trim(parnam) // "@" // trim(task)
      if (ind1>0 ) then
        write(ci, '(i1)') ind1
        parnam = trim(parnam) // "[" // ci // "]"
        if (ind2>0) then
          write(ci, '(i1)') ind1
          parnam = trim(parnam) // "[" // ci // "]"
        end if
      end if
    end if

    call lock(".","LOCK.PAR")
    open(unit=lupar, file="SASS.PAR", form="formatted", status="old", position="append")
    write(lupar,'(3a)') trim(parnam), "=", trim(adjustl(c))
    close(lupar)
    call unlock(".","LOCK.PAR")

  end subroutine sputc1

  !---------------------------------------------------------------
  subroutine sputc2(param, task, ind1, ind2, ind3, ind4, c, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: ind1, ind2, ind3, ind4
    integer, intent(out) :: ierr
    character(len=*), intent(in) :: param, task, c
    character(len=64) :: parnam
    character(len=2) :: ci
    integer :: unit

    ierr = 0 !*** but it can't be anything else. Could get rid of this argument.

    if (ind1>niter .or. ind2>niter .or. ind3>niter .or. ind4>niter) call wrerr("sputc2", 4, "Index out of range")

    parnam = trim(adjustl(param))
    if (len(trim(task))>0) then
      parnam = trim(parnam) // "@" // trim(task)
      if (ind1>0 ) then
        if (ind1<10) then
          write(ci, '(i1)') ind1
          parnam = trim(parnam) // "[" // ci(1:1) // "]"
        else
          write(ci, '(i2)') ind1
          parnam = trim(parnam) // "[" // ci // "]"
        end if
        if (ind2>0) then
          if (ind2<10) then
            write(ci, '(i1)') ind2
            parnam = trim(parnam) // "[" // ci(1:1) // "]"
          else
            write(ci, '(i2)') ind2
            parnam = trim(parnam) // "[" // ci // "]"
          end if
          if (ind3>0) then
            if (ind3<10) then
              write(ci, '(i1)') ind3
              parnam = trim(parnam) // "[" // ci(1:1) // "]"
            else
              write(ci, '(i2)') ind3
              parnam = trim(parnam) // "[" // ci // "]"
            end if
            if (ind4>0) then
              if (ind4<10) then
                write(ci, '(i1)') ind4
                parnam = trim(parnam) // "[" // ci(1:1) // "]"
              else
                write(ci, '(i2)') ind4
                parnam = trim(parnam) // "[" // ci // "]"
              end if
            end if
          end if
        end if
      end if
    end if

    call lock(".","LOCK.PAR")
    open(newunit=unit, file="SASS.PAR", form="formatted", status="old", position="append")
    write(unit,'(3a)') trim(parnam), "=", trim(adjustl(c))
    close(unit)
    call unlock(".","LOCK.PAR")

  end subroutine sputc2

  !---------------------------------------------------------------
  subroutine rgetc1(lupar, param, c, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar
    character(len=*), intent(in) :: param
    character(len=*), intent(out) :: c
    integer, intent(out) :: ierr

    integer, parameter :: istart=1, isize=1, lenmax=2048
    character(len=lenmax), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer :: icount

    call rgetn("C", lupar, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)
    c=ca(1)

  end subroutine rgetc1

  !---------------------------------------------------------------
  subroutine rgetc2(param, c, ierr)
    !---------------------------------------------------------------

    character(len=*), intent(in) :: param
    character(len=*), intent(out) :: c
    integer, intent(out) :: ierr

    integer, parameter :: istart=1, isize=1, lenmax=2048
    character(len=lenmax), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer :: icount

    call rgetn("C", 1, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)
    c=ca(1)

  end subroutine rgetc2

  !--------------------------------------------------------------
  subroutine rgetj1(lupar, param, i2, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar
    character(len=*), intent(in) :: param
    integer(kind=2), intent(out) :: i2
    integer, intent(out) :: ierr
 
    integer, parameter :: istart=1, isize=1
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer :: icount

    call rgetn("I2", lupar, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)
    i2=i2a(1)

  end subroutine rgetj1

  !--------------------------------------------------------------
  subroutine rgeti2(param, i2, ierr)
    !---------------------------------------------------------------

    character(len=*), intent(in) :: param
    integer(kind=2), intent(out) :: i2
    integer, intent(out) :: ierr
 
    integer, parameter :: istart=1, isize=1
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer :: icount

    call rgetn("I2", 1, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)
    i2=i2a(1)

  end subroutine rgeti2

  !---------------------------------------------------------------
  subroutine rgeti1(lupar, param, i4, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar
    character(len=*), intent(in) :: param
    integer, intent(out) :: i4
    integer, intent(out) :: ierr
 
    integer, parameter :: istart=1, isize=1
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer :: icount

    call rgetn("I4", lupar, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)
    i4 = i4a(1)

  end subroutine rgeti1

  !---------------------------------------------------------------
  subroutine rgetj2(param, i4, ierr)
    !---------------------------------------------------------------

    character(len=*), intent(in) :: param
    integer, intent(out) :: i4
    integer, intent(out) :: ierr
 
    integer, parameter :: istart=1, isize=1
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer :: icount

    call rgetn("I4", 1, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)
    i4 = i4a(1)

  end subroutine rgetj2

  !---------------------------------------------------------------
  subroutine rgetf1(lupar, param, r4, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar
    character(len=*), intent(in) :: param
    real(kind=4), intent(out) :: r4
    integer, intent(out) :: ierr
 
    integer, parameter :: istart=1, isize=1
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer :: icount

    call rgetn("R4", lupar, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)
    r4=r4a(1)

  end subroutine rgetf1

  !---------------------------------------------------------------
  subroutine rgetf2(param, r4, ierr)
    !---------------------------------------------------------------

    character(len=*), intent(in) :: param
    real(kind=4), intent(out) :: r4
    integer, intent(out) :: ierr
 
    integer, parameter :: istart=1, isize=1
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer :: icount

    call rgetn("R4", 1, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)
    r4=r4a(1)

  end subroutine rgetf2

  !---------------------------------------------------------------
  subroutine rgetd1(lupar, param, r8, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar
    character(len=*), intent(in) :: param
    real(kind=8), intent(out) :: r8
    integer, intent(out) :: ierr
 
    integer, parameter :: istart=1, isize=1
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer :: icount

    call rgetn("R8", lupar, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)
    r8=r8a(1)

  end subroutine rgetd1

  !---------------------------------------------------------------
  subroutine rgetd2(param, r8, ierr)
    !---------------------------------------------------------------

    character(len=*), intent(in) :: param
    real(kind=8), intent(out) :: r8
    integer, intent(out) :: ierr
 
    integer, parameter :: istart=1, isize=1
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer :: icount

    call rgetn("R8", 1, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)
    r8=r8a(1)

  end subroutine rgetd2

  !---------------------------------------------------------------
  subroutine rgetl1(lupar, param, l, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar
    character(len=*), intent(in) :: param
    logical, intent(out) :: l
    integer, intent(out) :: ierr
 
    integer, parameter :: istart=1, isize=1
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer :: icount

    call rgetn("L", lupar, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)
    l=la(1)

  end subroutine rgetl1

  !---------------------------------------------------------------
  subroutine rgetl2(param, l, ierr)
    !---------------------------------------------------------------

    character(len=*), intent(in) :: param
    logical, intent(out) :: l
    integer, intent(out) :: ierr
 
    integer, parameter :: istart=1, isize=1
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer :: icount

    call rgetn("L", 1, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)
    l=la(1)

  end subroutine rgetl2

  !---------------------------------------------------------------
  subroutine rgetca1(lupar, param, isize, ca, icount, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar, isize
    integer, intent(out) :: icount, ierr
    character(len=*), intent(in) :: param
 
    character(len=*), dimension(isize), intent(out) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer, parameter :: istart = 1

    call rgetn("C", lupar, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)

  end subroutine rgetca1

  !---------------------------------------------------------------
  subroutine rgetca2(param, isize, ca, icount, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: isize
    integer, intent(out) :: icount, ierr
    character(len=*), intent(in) :: param
 
    character(len=*), dimension(isize), intent(out) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer, parameter :: istart = 1

    call rgetn("C", 1, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)

  end subroutine rgetca2

  !---------------------------------------------------------------
  subroutine rgetia1(lupar, param, isize, i4a, icount, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar, isize
    integer, intent(out) :: icount, ierr
    character(len=*), intent(in) :: param
 
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize), intent(out) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer, parameter :: istart = 1

    call rgetn("I4", lupar, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)

  end subroutine rgetia1

  !---------------------------------------------------------------
  subroutine rgetja2(param, isize, i4a, icount, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: isize
    integer, intent(out) :: icount, ierr
    character(len=*), intent(in) :: param
 
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize), intent(out) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer, parameter :: istart = 1

    call rgetn("I4", 1, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)

  end subroutine rgetja2

  !---------------------------------------------------------------
  subroutine rgetja1(lupar, param, isize, i2a, icount, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar, isize
    integer, intent(out) :: icount, ierr
    character(len=*), intent(in) :: param
 
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize), intent(out) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer, parameter :: istart = 1

    call rgetn("I2", lupar, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)

  end subroutine rgetja1

  !---------------------------------------------------------------
  subroutine rgetia2(param, isize, i2a, icount, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: isize
    integer, intent(out) :: icount, ierr
    character(len=*), intent(in) :: param
 
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize), intent(out) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer, parameter :: istart = 1

    call rgetn("I2", 1, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)

  end subroutine rgetia2

  !---------------------------------------------------------------
  subroutine rgetfa1(lupar, param, isize, r4a, icount, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar, isize
    integer, intent(out) :: icount, ierr
    character(len=*), intent(in) :: param
 
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize), intent(out) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer, parameter :: istart = 1

    call rgetn("R4", lupar, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)

  end subroutine rgetfa1

  !---------------------------------------------------------------
  subroutine rgetfa2(param, isize, r4a, icount, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: isize
    integer, intent(out) :: icount, ierr
    character(len=*), intent(in) :: param
 
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize), intent(out) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer, parameter :: istart = 1

    call rgetn("R4", 1, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)

  end subroutine rgetfa2

  !---------------------------------------------------------------
  subroutine rgetda1(lupar, param, isize, r8a, icount, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar, isize
    integer, intent(out) :: icount, ierr
    character(len=*), intent(in) :: param
 
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize), intent(out) :: r8a

    integer, parameter :: istart = 1

    call rgetn("R8", lupar, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)

  end subroutine rgetda1

  !---------------------------------------------------------------
  subroutine rgetda2(param, isize, r8a, icount, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: isize
    integer, intent(out) :: icount, ierr
    character(len=*), intent(in) :: param
 
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize), intent(out) :: r8a

    integer, parameter :: istart = 1

    call rgetn("R8", 1, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)

  end subroutine rgetda2

  !---------------------------------------------------------------
  subroutine rgetla1(lupar, param, isize, la, icount, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar, isize
    integer, intent(out) :: icount, ierr
    character(len=*), intent(in) :: param
 
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize), intent(out) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer, parameter :: istart = 1

    call rgetn("L", lupar, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)

  end subroutine rgetla1

  !---------------------------------------------------------------
  subroutine rgetla2(param, isize, la, icount, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: isize
    integer, intent(out) :: icount, ierr
    character(len=*), intent(in) :: param
 
    character(len=1), dimension(isize) :: ca
    integer(kind=2), dimension(isize) :: i2a
    integer(kind=4), dimension(isize) :: i4a
    logical, dimension(isize), intent(out) :: la
    real(kind=4), dimension(isize) :: r4a
    real(kind=8), dimension(isize) :: r8a

    integer, parameter :: istart = 1

    call rgetn("L", 1, param, istart, isize, ca, la, &
      i2a, i4a, r4a, r8a, icount, ierr)

  end subroutine rgetla2

  !---------------------------------------------------------------
  subroutine sgetc1(lupar, param, task, ind1, ind2, cval, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: lupar, ind1, ind2
    integer, intent(out) :: ierr
    character(len=*), intent(in) :: param, task
    character(len=*), intent(out) :: cval

    character(len=2048) :: parline
    character(len=64) :: parnam
    character(len=1) :: ci
    integer :: read_stat, expos, eqpos
    logical :: found

    parnam=adjustl(param)
    if (len_trim(task)>0) then
      parnam = trim(parnam) // "@" // trim(task)
      if (ind1>0 ) then
        write(ci, '(i1)') ind1
        parnam = trim(parnam) // "[" // ci // "]"
        if (ind2>0) then
          write(ci, '(i1)') ind1
          parnam = trim(parnam) // "[" // ci // "]"
        end if
      end if
    end if

    call lock(".","LOCK.PAR")
    open(unit=lupar, file="SASS.PAR", form="formatted", status="old")
    found = .false.
    do
      read (lupar,'(a)', iostat=read_stat) parline
      if (read_stat /= 0) exit
      expos = index(parline,"!")
      if (expos>0) parline=parline(1:expos-1)
      eqpos = index(parline, "=")
      if (eqpos>1 .and. trim(parnam)==trim(adjustl(parline(1:eqpos-1)))) then
        cval=adjustl(parline(eqpos+1:))
        found = .true.
      end if
    end do
    close(lupar)
    call unlock(".","LOCK.PAR")

    if (found) then
      ierr = 0
    else
      ierr = 1
    end if

  end subroutine sgetc1

  !---------------------------------------------------------------
  subroutine sgetc2(param, task, ind1, ind2, ind3, ind4, cval, ierr)
    !---------------------------------------------------------------

    integer, intent(in) :: ind1, ind2, ind3, ind4
    integer, intent(out) :: ierr
    character(len=*), intent(in) :: param, task
    character(len=*), intent(out) :: cval

    character(len=2048) :: parline
    character(len=64) :: parnam
    character(len=2) :: ci
    integer :: unit, read_stat, expos, eqpos
    logical :: found, cont

    if (ind1>niter .or. ind2>niter .or. ind3>niter .or. ind4>niter) call wrerr("sgetc2", 4, "Index out of range")

    parnam = trim(adjustl(param))
    if (len(trim(task))>0) then
      parnam = trim(parnam) // "@" // trim(task)
      if (ind1>0 ) then
        if (ind1<10) then
          write(ci, '(i1)') ind1
          parnam = trim(parnam) // "[" // ci(1:1) // "]"
        else
          write(ci, '(i2)') ind1
          parnam = trim(parnam) // "[" // ci // "]"
        end if
        if (ind2>0) then
          if (ind2<10) then
            write(ci, '(i1)') ind2
            parnam = trim(parnam) // "[" // ci(1:1) // "]"
          else
            write(ci, '(i2)') ind2
            parnam = trim(parnam) // "[" // ci // "]"
          end if
          if (ind3>0) then
            if (ind3<10) then
              write(ci, '(i1)') ind3
              parnam = trim(parnam) // "[" // ci(1:1) // "]"
            else
              write(ci, '(i2)') ind3
              parnam = trim(parnam) // "[" // ci // "]"
            end if
            if (ind4>0) then
              if (ind4<10) then
                write(ci, '(i1)') ind4
                parnam = trim(parnam) // "[" // ci(1:1) // "]"
              else
                write(ci, '(i2)') ind4
                parnam = trim(parnam) // "[" // ci // "]"
              end if
            end if
          end if
        end if
      end if
    end if

    call lock(".","LOCK.PAR")
    open(newunit=unit, file="SASS.PAR", form="formatted", status="old")
    found = .false.; cont = .false.; cval = ""
    do
      read (unit,'(a)', iostat=read_stat) parline
      if (read_stat /= 0) exit
      expos = index(parline,"!")
      if (expos>0) parline=parline(1:expos-1)

      ! append continuation lines
      if (found .and. cont) then
        if (index(parline, "=")==0) then
          cval = trim(cval) // " " // trim(parline)
        else
          cont = .false.
        end if
      end if

      eqpos = index(parline, "=")
      if (eqpos>1 .and. trim(parnam)==trim(adjustl(parline(1:eqpos-1)))) then
        cval=adjustl(parline(eqpos+1:))
        found = .true.; cont = .true.
      end if

    end do

    close(unit)
    call unlock(".","LOCK.PAR")

    if (found) then
      ierr = 0
    else
      ierr = 1
    end if

  end subroutine sgetc2

  !---------------------------------------------------------------------------
  subroutine pgeti(parfile, param, ival, ierr)
    !---------------------------------------------------------------------------

    character(len=*), intent(in) :: parfile
    character(len=*), intent(in) :: param
    integer, intent(out) :: ival, ierr

    integer, parameter :: clen=256
    integer :: unit, expos, eqpos, read_stat

    character(len=clen) line, parname, parval

    ierr = 0 !*** but it can't be anything else. Could get rid of this argument.

    open(newunit=unit, file=parfile, status="old", form="formatted")

    do
    
      read (unit, '(a)', iostat=read_stat) line
      if (read_stat /= 0) exit
   
      expos=index(line, "!")
      if (expos>0) line=line(1:expos-1)
      eqpos = index(line, "=")
      if (eqpos>1) then
        parname=trim(adjustl(line(1:eqpos-1)))
        parval =trim(adjustl(line(eqpos+1:)))
        if (trim(parname)==trim(param)) read (parval,*) ival
      end if

    end do
    close(unit)

  end subroutine pgeti

  !---------------------------------------------------------------------------
  subroutine pgetc(parfile, param, cval, ierr)
    !---------------------------------------------------------------------------

    character(len=*), intent(in) :: parfile
    character(len=*), intent(in) :: param
    character(len=*), intent(out) :: cval
    integer, intent(out) :: ierr

    integer, parameter :: clen=256
    integer :: unit, expos, eqpos, read_stat
    logical :: found, cont

    character(len=clen) line, parname, parval

    ierr = 0 !*** but it can't be anything else. Could get rid of this argument.

    open(newunit=unit, file=parfile, status="old", form="formatted")

    found=.false.; cont = .false.; cval=""
    do
    
      read (unit, '(a)', iostat=read_stat) line
      if (read_stat /= 0) exit
   
      expos=index(line, "!")
      if (expos>0) line=line(1:expos-1)

      ! append continuation lines
      if (found .and. cont) then
        if (index(line, "=")==0) then
          cval = trim(cval) // " " // trim(line)
        else
          cont = .false.
        end if
      end if

      eqpos = index(line, "=")
      if (eqpos>1) then
        parname=trim(adjustl(line(1:eqpos-1)))
        parval =trim(adjustl(line(eqpos+1:)))
        if (trim(parname)==trim(param)) then
          found = .true.; cont = .true.
          cval=trim(parval)
        end if
      end if
   
    end do

    close(unit)
    if (found .and. len_trim(cval)==0) call wrerr("pgetc", 1, "String parameter not set")

  end subroutine pgetc

  !---------------------------------------------------------------------------
  subroutine rgetn(cid, lupar, param, istart, isize, ca, la, &
    i2a , i4a, r4a, r8a, icnt, ierr)
    !---------------------------------------------------------------------------

    ! gets parameter values from sequential ascii file (replaces old ROSAT/Abrixas scheme)
    ! syntax:
    ! <param> = <val1> [, <val2] ... [, valmax]
    ! blanks before and after <param> and <valn> are ignored
    ! linelen is max line length, no continuation lines
    ! max number of array elements is given by isize
    ! maxbuff is max number of lines
    ! if a parameter appears multiple times, the last instance will be used
    ! lines with don't contain a "=" are treated as comment lines
    ! everything after "!" is be treated as comment
    ! "=" and "," are not permitted in parameter names or values
    ! the parameter file is copied to an internal buffer in the first rget* call
    ! rereading of the parameter file can be enforced by passing a negativ lupar
    !   value in the rget* callx

    ! call_var.          type I/O description
    ! CA                  C   O   character array   \
    ! CID                 C   I  identifier for data type
    ! I2A                 I2    O Integer*2 array    |
    ! I4A                 I4    O Integer*4 array    | buffers to
    ! ICNT                I4    O number of values found
    ! IERR                I4    O = 0 no error.
    !                             = 1 Empty parameter value (pipeline).
    !                             = 2 Error getting parameter (non-pipeline).
    !                             = 3 IO error. (*** doesn't seem ever to be set?)
    !                             = 4 MAXBUFF too small, or loop index out of range, or parameter not found, or cannot read, or parameter type not recognized. (** should have separate values?)
    ! ISIZE               I4  I   size of buffer for values
    ! ISTART              I4  I   first value to be taken
    ! LA                  L4  I   Logical   array    | receive
    ! LUPAR               I4  I logical number for parameter file
    ! PARAM               C   I  parameter name to be searched for
    ! R4A                 R4  I/? Real*4 array       | found values
    ! R8A                 R8  ?/? Real*8 array       |

    integer, intent(in) :: lupar, istart, isize
    character(len=*), intent(in) :: param, cid
 
    character(len=*), dimension(isize), intent(out) :: ca
    integer(kind=2), dimension(isize), intent(out) :: i2a
    integer(kind=4), dimension(isize), intent(out) :: i4a
    integer(kind=8) :: it8
    real(kind=4), dimension(isize), intent(out) :: r4a
    real(kind=8), dimension(isize), intent(out) :: r8a
    logical, dimension(isize), intent(out) :: la

    integer, intent(out) :: icnt, ierr

    integer, parameter :: linelen=8192, maxbuff=2048, clen=256
    character(len=linelen), dimension(maxbuff) :: parbuff
    character(len=linelen) :: parline, parline1, cval, lline
    character(len=clen), dimension(isize) :: carr
    character(len=clen) :: tasknam
    character(len=clen) :: parnam, parnam0, parnam1, parnam2, parnam3, parnam02, parnam023, parnam003
    character(len=clen) :: lcparnam, lcparnam0, lcparnam1, parstring
    character(len=64) :: cvers
    character(len=32) :: date, chost
    character(len=8) :: par_int
    character(len=2) :: cind, cll

    character(len=64) :: version ! version.for function (created by eSASS make)

    logical :: first = .true., found = .false.
   
    integer :: i, ind, ind1, ind2, ind3, ind4, read_stat, eqpos, expos, compos, istat, len, &
      sp1, sp2, mn, ll, l1, l2, itry, maxline, maxline1, lu, readstat, status, iostat

    character(linelen) :: str_wrchatty

    save parbuff, first, ind1, ind2, ind3, ind4

    ierr  = 0

    ! check environment variable  SASS_PARINT, if set to 'PIPELINE',
    ! use SASS.PAR file, otherwise call PIL interface (the default).

    call get_environment_variable('SASS_PARINT', par_int)
    if (par_int .ne. 'PIPELINE' .and. par_int .ne. 'pipeline' ) then
      ! write (*,*) '*********** USE PIL ********'
      call rgetn_pil(cid, param, istart, isize, ca, la, i2a, i4a, r4a, r8a, icnt, ierr)
    else
      ! write (*,*) '******* USE SASS.PAR ********'

      ! read parameters into buffer if first call or lupar<0
      if (first .or. lupar<0) then

        lu=abs(lupar)
        call lock(".","LOCK.PAR")
        open(unit=lu,file="SASS.PAR", form="formatted", status="old")
        ! read par file into buffer, handle comments and continuation lines
        i=0
        do while (i<=maxbuff)
          read (lu,'(a)', iostat=read_stat) parline
          if (read_stat /= 0) exit
          expos = index(parline,"!")
          if (expos>0) parline=parline(1:expos-1)
          eqpos = index(parline, "=")
          if (eqpos/=0) then
            i=i+1
            parbuff(i) = ""
          end if
          if (i>=1) parbuff(i) = trim(parbuff(i)) // " " // trim(parline)
        end do
        close(lu)
        call unlock(".","LOCK.PAR")
        maxline = i
        if (maxline > maxbuff) then
          ierr = 4
          call wrerr("rgetn", ierr, "MAXBUFF too small")
          return
        end if
      end if

      ! expand first loop
      maxline1=maxline
      do i=1, maxline
        lline = parbuff(i)
        if (index(lline,"@")>1) then
          sp1=index(lline,"[")
          sp2=index(lline,"]")
          mn=index(lline,"-")
          if (sp1>0.and.mn>sp1.and.sp2>mn) then
            if (mn-sp1==2) then
              read(lline(sp1+1:sp1+1),'(i1)') l1
            else if (mn-sp1==3) then
              read(lline(sp1+1:sp1+2),'(i2)') l1
            else
              call wrerr("rgetn", 4, "Can't parse loop range")
            end if
            if (sp2-mn==2) then
              read(lline(sp2-1:sp2-1),'(i1)') l2
            else if (sp2-mn==3) then
              read(lline(sp2-2:sp2-1),'(i2)') l2
            else
              call wrerr("rgetn", 4, "Can't parse loop range")
            end if
            do ll=l1,l2
              maxline1=maxline1+1
              if (ll<10) then
                write(cll,'(i1)') ll
              else
                write(cll,'(i2)') ll
              end if
              parbuff(maxline1)=lline(:sp1) // trim(adjustl(cll)) // lline(sp2:)
            end do
          end if
        end if
      end do
      maxline = maxline1

      ! get task name
      call get_command_argument(0, tasknam)
      ! fix for c program ftfindhotpix
      !    if (len_trim(tasknam)==0) tasknam="ftfindhotpix"

      if (tasknam=="proc") then
        ! there are no loop indices if called from proc
        ind1=0; ind2=0; ind3=0; ind4=0
      else
        ! get loop indices from command line
        call get_command_argument(1, cind, len, istat)
        if (istat==0 .and. len<=2) then
          read(cind,*,iostat=iostat) ind1
        else
          ind1=1
        end if
        if (iostat/=0) ind1=1

        call get_command_argument(2, cind, len, istat)
        if (istat==0 .and. len<=2) then
          read(cind,*,iostat=iostat) ind2
        else
          ind2=1
        end if
        if (iostat/=0) ind2=1

        call get_command_argument(3, cind, len, istat)
        if (istat==0 .and. len<=2) then
          read(cind,*,iostat=iostat) ind3
        else
          ind3=1
        end if
        if (iostat/=0) ind3=1

        call get_command_argument(4, cind, len, istat)
        if (istat==0 .and. len<=2) then
          read(cind,*,iostat=iostat) ind4
        else
          ind4=1
        end if
        if (ind1>niter .or. ind2>niter .or. ind3>niter .or. ind4>niter) then
          ierr=4
          call wrerr("rgetn",ierr,"Loop index out of range")
        end if
        if (iostat/=0) ind4=1

        ind1 = max(1,ind1)
        ind2 = max(1,ind2)
        ind3 = max(1,ind3)
        ind4 = max(1,ind4)
      end if 

      ! find parameter, copy value into cval
      ! leading and trailing blanks are ignored; indices in the range 1..niter are supported;
      ! the program looks, in this order, for parameter names of any of the following forms;
      ! the most specific (last) form matching the array indices ind1/ind2/ind3/ind4 will be
      ! selected:

      ! 1) <param>
      ! 2) <param>@<taskname>
      ! 3) <param>@<taskname>[ind1]
      ! 4) <param>@<taskname>[ind1][ind2]
      ! 5) <param>@<taskname>[ind1][ind2][ind3]
      ! 6) <param>@<taskname>[ind1][ind2][ind3][ind4]

      parnam = trim(adjustl(param))

      found = .false.
      do itry=1,6

        if (itry==2) then
          parnam = trim(parnam) // "@" // trim(tasknam)
          parnam0 = parnam
        end if
        if (itry==3) then
          if (ind1<10) then
            write(cind,'(i1)') ind1
            parnam = trim(parnam) // "[" // cind(1:1) // "]"
          else
            write(cind,'(i2)') ind1
            parnam = trim(parnam) // "[" // cind // "]"
          end if
          parnam1=parnam
        end if
        if (itry==4) then
          if (ind2<10) then
            write(cind,'(i1)') ind2
            parnam = trim(parnam1) // "[" // cind(1:1) // "]"
          else
            write(cind,'(i2)') ind2
            parnam = trim(parnam1) // "[" // cind // "]"
          end if
          parnam2=parnam
        end if
        if (itry==5) then
          if (ind3<10) then
            write(cind,'(i1)') ind3
            parnam = trim(parnam2) // "[" // cind(1:1) // "]"
          else
            write(cind,'(i2)') ind3
            parnam = trim(parnam2) // "[" // cind // "]"
          end if
          parnam3=parnam
        end if
        if (itry==6) then
          if (ind4<10) then
            write(cind,'(i1)') ind4
            parnam = trim(parnam3) // "[" // cind(1:1) // "]"
          else
            write(cind,'(i2)') ind4
            parnam = trim(parnam3) // "[" // cind // "]"
          end if
        end if
 
        call lcase(parnam, lcparnam)

        do i=1, maxline
          parline=parbuff(i)
          eqpos = index(parline, "=")
          if (eqpos>1) then
            call lcase(adjustl(parline(1:eqpos-1)), lcparnam1)
            if (lcparnam==lcparnam1) then
              cval=parline(eqpos+1:)
              found = .true.
            end if
          end if
        end do

        ! don't test for loop parameters if called from proc
        if (tasknam=="proc") exit
      end do

      if (.not. found) then
        ierr=4
        !      call lcase(parnam0, lcparnam0)
        call wrerr("rgetn", ierr, "Parameter not found: " //  trim(parnam))
        return
      end if

      !end if

      ! copy array components into carr, remove leading and trailing blanks
!      if (len_trim(adjustl(cval))==0) then
      if (len_trim(adjustl(cval))<1) then
        ! Empty value string: array parameter set to length 0;
        !   single parameter set to "", .false., 0, respectively.
        icnt=0
        ca=""; la=.false.; i2a=0; i4a=0; r4a=0.; r8a=0.
        ierr=1
        call wrerr("rgetn", ierr, "Empty parameter value")
  return
      else
        ! length 1 to isize
        do i=1,isize
          compos=index(cval, ",")
          if (compos>0) then
            carr(i)=trim(adjustl(cval(1:compos-1)))
            cval=cval(compos+1:)
          else
            carr(i)=trim(adjustl(cval))
            exit
          end if
        end do
        icnt=i-istart+1
      end if
   
      ! decode parameter value, depending on type
      do ind=1,icnt

        select case (cid)
          case('C', 'c')
            ca(ind)=carr(istart+ind-1)
          case ('L', 'l')
            if (carr(istart+ind-1)(1:1)=="F" .or. carr(istart+ind-1)(1:1)=="N" .or. &
              carr(istart+ind-1)(1:1)=="f" .or. carr(istart+ind-1)(1:1)=="n") then
              la(ind)=.false.
            else
              la(ind)=.true.
            end if
          case('I2','i2')
            read(carr(istart+ind-1),*,iostat=readstat) i2a(ind)
            if (readstat>0) then
              ierr=4
              call wrerr("rgetn",ierr,"Cannot read " // trim(carr(istart+ind-1)))
            end if
          case('I4','i4')
            read(carr(istart+ind-1),*,iostat=readstat) i4a(ind)
            if (readstat>0) then
              ierr=4
              call wrerr("rgetn",ierr,"Cannot read " // trim(carr(istart+ind-1)))
            end if
          case('R4','r4')
            read(carr(istart+ind-1),*,iostat=readstat) r4a(ind)
            if (readstat>0) then
              ierr=4
              call wrerr("rgetn",ierr,"Cannot read " // trim(carr(istart+ind-1)))
            end if
          case('R8','r8')
            read(carr(istart+ind-1),*,iostat=readstat) r8a(ind)
            if (readstat>0) then
              ierr=4
              call wrerr("rgetn",ierr,"Cannot read " // trim(carr(istart+ind-1)))
            end if
          case default
            ierr=4
            call wrerr("rgetn",ierr,'Param. type '//cid//' not supported')
            return
        end select
      end do
                       
    end if

    ! update sass_history string (GL+HB):
    if (ierr == 0) then
      if (len_trim(sass_history)==0) then
        it8=time8()
        call ctime(it8, date)
        call hostnm(chost, status)
        call get_command_argument(0, tasknam)
        cvers = version()
        sass_history = " **** DATE:" // trim(date(5:)) // &
          " HOST:" // trim(chost) // &
          " TASK:" // trim(tasknam) // " " // trim(cvers) // &
          " PARAMS:" // trim(param) // "="
      else
        sass_history = trim(sass_history) // " " // trim(param) // '='
      end if

      str_wrchatty = trim(param) // ":"

      if (icnt>1)  sass_history = trim(sass_history) // '"'
      select case (cid)
        case('C', 'c')
          sass_history = trim(sass_history) // adjustl(trim(ca(1)))
          str_wrchatty = trim(str_wrchatty) // " " // adjustl(trim(ca(1)))
          do ind=2,icnt
            sass_history = trim(sass_history) // ' ' // adjustl(trim(ca(ind)))
            str_wrchatty = trim(str_wrchatty) // " " // adjustl(trim(ca(ind)))
          end do
        case('L', 'l')
          write(parstring,*) la(1)
          sass_history = trim(sass_history) // adjustl(trim(parstring))
          str_wrchatty = trim(str_wrchatty) // " " // adjustl(trim(parstring))
          do ind=2,icnt
            write(parstring,*) la(ind)
            sass_history = trim(sass_history) // ' ' // adjustl(trim(parstring))
            str_wrchatty = trim(str_wrchatty) // " " // adjustl(trim(parstring))
          end do
        case('I2','i2')
          write(parstring,*) i2a(1)
          sass_history = trim(sass_history) // adjustl(trim(parstring))
          str_wrchatty = trim(str_wrchatty) // " " // adjustl(trim(parstring))
          do ind=2,icnt
            write(parstring,*) i2a(ind)
            sass_history = trim(sass_history) // ' ' // adjustl(trim(parstring))
            str_wrchatty = trim(str_wrchatty) // " " // adjustl(trim(parstring))
          end do
        case('I4','i4')
          write(parstring,*) i4a(1)
          sass_history = trim(sass_history) // adjustl(trim(parstring))
          str_wrchatty = trim(str_wrchatty) // " " // adjustl(trim(parstring))
          do ind=2,icnt
            write(parstring,*) i4a(ind)
            sass_history = trim(sass_history) // ' ' // adjustl(trim(parstring))
            str_wrchatty = trim(str_wrchatty) // " " // adjustl(trim(parstring))
          end do
        case('R4','r4')
          write(parstring,*) r4a(1)
          sass_history = trim(sass_history) // adjustl(trim(parstring))
          str_wrchatty = trim(str_wrchatty) // " " // adjustl(trim(parstring))
          do ind=2,icnt
            write(parstring,*) r4a(ind)
            sass_history = trim(sass_history) // ' ' // adjustl(trim(parstring))
            str_wrchatty = trim(str_wrchatty) // " " // adjustl(trim(parstring))
          end do
        case('R8','r8')
          write(parstring,*) r8a(1)
          sass_history = trim(sass_history) // adjustl(trim(parstring))
          str_wrchatty = trim(str_wrchatty) // " " // adjustl(trim(parstring))
          do ind=2,icnt
            write(parstring,*) r8a(ind)
            sass_history = trim(sass_history) // ' ' // adjustl(trim(parstring))
            str_wrchatty = trim(str_wrchatty) // " " // adjustl(trim(parstring))
          end do
      end select
      if (icnt>1) sass_history =  trim(sass_history) // '"'

    endif

    call wrchatty("read params", trim(str_wrchatty))

  ! 1) <param>
  ! 2) <param>@<taskname>                            *
  ! 3) <param>@<taskname>[ind1]                      *
  ! 4) <param>@<taskname>[][ind2]
  ! 5) <param>@<taskname>[][][ind3]
  ! 6) <param>@<taskname>[][][][ind4]
  ! 7) <param>@<taskname>[ind1][ind2]                *
  ! 8) <param>@<taskname>[][ind2][ind3]
  ! 9) <param>@<taskname>[][][ind3][ind4]
  ! 10) <param>@<taskname>[ind1][ind2][ind3]         *
  ! 11) <param>@<taskname>[][ind2][ind3][ind4]
  ! 12) <param>@<taskname>[ind1][ind2][ind3][ind4]   *
  !
  !     parnam = trim(adjustl(param))
  !
  !     found = .false.
  !     do itry=1,12
  !
  !      if (itry==2) then
  !         parnam = trim(parnam) // "@" // trim(tasknam)
  !         parnam0 = parnam
  !      end if
  !      if (itry==3) then
  !        if (ind1<10) then
  !          write(cind,'(i1)') ind1
  !          parnam = trim(parnam) // "[" // cind(1:1) // "]"
  !        else
  !          write(cind,'(i2)') ind1
  !          parnam = trim(parnam) // "[" // cind // "]"
  !        end if
  !        parnam1=parnam
  !      end if
  !      if (itry==4) then
  !        if (ind2<10) then
  !          write(cind,'(i1)') ind2
  !          parnam = trim(parnam0) // "[][" // cind(1:1) // "]"
  !        else
  !          write(cind,'(i2)') ind2
  !          parnam = trim(parnam0) // "[][" // cind // "]"
  !        end if
  !        parnam02=parnam
  !      end if
  !      if (itry==5) then
  !        if (ind3<10) then
  !          write(cind,'(i1)') ind3
  !          parnam = trim(parnam0) // "[][][" // cind(1:1) // "]"
  !        else
  !          write(cind,'(i2)') ind3
  !          parnam = trim(parnam0) // "[][][" // cind // "]"
  !        end if
  !        parnam003=parnam
  !      end if
  !      if (itry==6) then
  !        if (ind4<10) then
  !          write(cind,'(i1)') ind4
  !          parnam = trim(parnam0) // "[][][][" // cind(1:1) // "]"
  !        else
  !          write(cind,'(i2)') ind4
  !          parnam = trim(parnam0) // "[][][][" // cind // "]"
  !        end if
  !      end if
  !      if (itry==7) then
  !        if (ind2<10) then
  !          write(cind,'(i1)') ind2
  !          parnam = trim(parnam1) // "[" // cind(1:1) // "]"
  !        else
  !          write(cind,'(i2)') ind2
  !          parnam = trim(parnam1) // "[" // cind // "]"
  !        end if
  !        parnam2=parnam
  !      end if
  !      if (itry==8) then
  !        if (ind3<10) then
  !          write(cind,'(i1)') ind3
  !          parnam = trim(parnam02) // "[" // cind(1:1) // "]"
  !        else
  !          write(cind,'(i2)') ind3
  !          parnam = trim(parnam02) // "[" // cind // "]"
  !        end if
  !        parnam023=parnam
  !      end if
  !      if (itry==9) then
  !        if (ind4<10) then
  !          write(cind,'(i1)') ind4
  !          parnam = trim(parnam003) // "[" // cind(1:1) // "]"
  !        else
  !          write(cind,'(i2)') ind4
  !          parnam = trim(parnam003) // "[" // cind // "]"
  !        end if
  !      end if
  !      if (itry==10) then
  !        if (ind3<10) then
  !          write(cind,'(i1)') ind3
  !          parnam = trim(parnam2) // "[" // cind(1:1) // "]"
  !        else
  !          write(cind,'(i2)') ind3
  !          parnam = trim(parnam2) // "[" // cind // "]"
  !        end if
  !      end if
  !      if (itry==11) then
  !        if (ind4<10) then
  !          write(cind,'(i1)') ind4
  !          parnam = trim(parnam023) // "[" // cind(1:1) // "]"
  !        else
  !          write(cind,'(i2)') ind4
  !          parnam = trim(parnam023) // "[" // cind // "]"
  !        end if
  !      end if
  !      if (itry==12) then
  !        if (ind4<10) then
  !          write(cind,'(i1)') ind4
  !          parnam = trim(parnam3) // "[" // cind(1:1) // "]"
  !        else
  !          write(cind,'(i2)') ind4
  !          parnam = trim(parnam3) // "[" // cind // "]"
  !        end if
  !      end if

  end subroutine rgetn

  !---------------------------------------------------------------------------
  subroutine lcase(str, lcstr)
    !---------------------------------------------------------------------------

    character(len=*), intent(in) :: str
    character(len=*), intent(out) :: lcstr
    character(len=1) :: ci
    integer :: length, i

    lcstr=""
    if (ichar('A')==65) then
      length = min(64,len_trim(str))
      do i = 1, length
        ci = str(i:i)
        if (ichar(ci)>=65 .and. ichar(ci)<=90) then
          lcstr(i:i)=char(ichar(ci)+32)
        else
          lcstr(i:)=str(i:i)
        end if
      end do
    else
      call wrerr("lcase",2,"Not ASCII character set, parameter names are case sensitive")
    end if

  end subroutine lcase

  !---------------------------------------------------------------------------
  SUBROUTINE RGETN_PIL(CID,PARAM,ISTART,ISIZE,CA,LA,I2A,I4A,R4A,R8A,ICNT,IERR)
    !---------------------------------------------------------------------------

    !  gets parameter values from PIL, equivalent to routine RGETN
    !
    !  for istart=isize=1 call PIL routine to get scalar value and
    !  depending on value of CID, fill 1st element of one of
    !  the arrays CA, BA,...,R8A
    !  If ISTART > 1 or ISIZE > 1, call PIL array reading routines,
    !  determine number of array elements J, the
    !  elements 1:ISTART-1 of the parameter array read are ignored
    !  ICNT = J-ISTART+1
    !  Fill  1:ICNT of relevant array, return ICNT
    !
    !  author : GL       date: 29-03-2010
    !   call_var.          type I/O description
    !  CA                  C   O   character array   \
    !  CID                 C   I  identifier for data type
    !  I2A                 I2    O Integer*2 array    |
    !  I4A                 I4    O Integer*4 array    | buffers to
    !  ICNT                I4    O number of values found
    !  IERR                I4    O = 0 no error
    !                              = 2 error
    !  ISIZE               I4  I   size of buffer for values
    !  ISTART              I4  I   first value to be taken
    !  LA                  L4  I   Logical   array    | receive
    !  LUPAR               I4  I logical number for parameter file
    !  PARAM               C   I  parameter name to be searched for
    !  R4A                 R4  I/? Real*4 array       | found values
    !  R8A                 R8  ?/? Real*8 array       |

    use pil_f90_api
             
    CHARACTER*(*) CA(ISIZE) ,PARAM, CID
    CHARACTER*2048  CH_BUF
    !      CHARACTER*128  CH_BUF
    INTEGER*4     ISIZE
    INTEGER*2     I2A(ISIZE)
    INTEGER*4     I4A(ISIZE) ,ISTART ,ICNT ,LUPAR ,IERR
    LOGICAL       LA(ISIZE)
    REAL*4        R4A(ISIZE),R4ARR(ISIZE),r4val
    REAL*8        R8A(ISIZE),R8ARR(ISIZE)
    INTEGER*4     IND,IVAL,II,J
    INTEGER       ISTAT
    INTEGER*4     IARR(ISIZE)
    integer :: lenc, lenb, lend
    CHARACTER*9   RTNAME

    DATA          RTNAME /'rgetn_pil'/

    ierr  = 0
    istat = 0
    !      istat = PILINIT()
    !      IF (istat .ne. 0) then
    !        ierr=2
    !        CALL wrerr(rtname,ierr,'Error initializing PIL')
    !      ELSE
    IF (istart .eq. 1 .and. isize .eq. 1) THEN
      icnt=1
      ind=1
      SELECT CASE(CID)
        CASE('C','c')
          istat = PILgetstring(trim(param),ca(ind))
          !write(*,*) RTNAME,' ', ca(ind)
        CASE('L','l')
          istat = PILgetbool(trim(param),ival)
          if (ival .eq. 0) then
            la(ind)=.false.
          else
            la(ind)=.true.
          endif
        CASE('I2','i2')
          istat = PILgetint(trim(param),ival)
          i2a(ind)=ival
        CASE('I4','i4')
          istat = PILgetint(trim(param),i4a(ind))
        CASE('R4','r4')
          istat = PILgetreal4(trim(param),r4a(ind))
        CASE('R8','r8')
          istat = PILgetreal(trim(param),r8a(ind))
        CASE DEFAULT
          istat=1
          CALL wrerr(rtname,1,'Param. type '//cid//' not supported')
      END SELECT

    ELSE IF (istart .gt. 0 .and. isize .gt. 0) THEN
      SELECT CASE(CID)
        CASE('C','c')
          istat = PILgetstring(trim(param),ch_buf)
          ii=0
          lenc = len(ca(1))
          do
            ch_buf = adjustl(ch_buf)
            lenb = len_trim(ch_buf)
            lend = index(ch_buf," ")-1
            if (lend<0) lend=lenb
            if (lend>0) then
              ii=ii+1
              if (ii>isize-istart+1) &
                call wrerr(rtname,4,'Too many elements in param. '//param)
              if (ii>=istart) then
                if (lend>lenc) &
                  call wrerr(rtname,4,'Element truncated in param. '//param)
                ca(ii-istart+1)=ch_buf(:lend)
              end if
            end if
            if (lend+2<=lenb) then
              ch_buf = ch_buf(lend+2:)
            else
              exit
            end if
          end do
          icnt=max(0,ii-istart+1)

        !!            do j=1, isize
        !!              ind=index(ch_buf,' ')
        !!              if (ind .le. 1) then
        !               last element reached
        !!                ind = len(trim(ch_buf))+1
        !!              endif
        !             ignore elements (1:istart-1):
        !!              if (j .ge. istart) then
        !!                ii=ii+1
        !!                if (ii .le. isize) then
        !!                  ca(ii)=ch_buf(1:ind-1)
                !write(*,*) RTNAME,' ', ca(ii)
        !!                else
        !!                  istat=1
        !!                  CALL wrerr(rtname,1,'Too many elements in param. '//param)
        !!                endif
        !!              endif
        !!              ch_buf=ch_buf(ind+1:)
        !!              if (len(trim(ch_buf)) .le. 1) then
        !               no more elements in ch_buf
        !!                exit
        !!              endif
        !!            enddo
        !!            icnt=j-istart+1
            
        CASE('L','l')
          istat = PILgetstring(trim(param),ch_buf)
          ii=0
          do j=1, isize
            ind=index(ch_buf,' ')
            if (ind .le. 1) then
              !               last element reached
              ind = len(trim(ch_buf))
            endif
            !             ignore elements (1:istart-1) :
            if (j .ge. istart) then
              ii=ii+1
              if (ii .le. isize) then
                SELECT CASE(ch_buf(1:1))
                  CASE('Y','y','T','t')
                    la(ii)=.true.
                  CASE('N','n','F','f')
                    la(ii)=.false.
                  CASE DEFAULT
                    istat = 1
                    CALL wrerr(rtname,1,'Not a valid boolean value')
                END SELECT
              else
                istat=1
                CALL wrerr(rtname,1,'Too many elements in param. '//param)
              endif
            endif
            ch_buf=ch_buf(ind+1:)
            if (len(trim(ch_buf)) .lt. 1) then
              !               no more elements in ch_buf
              exit
            endif 
          enddo
          icnt=j-istart+1

        CASE('I2','i2')
          icnt=isize
          istat = PILgetintvarvector(trim(param),icnt,iarr(1))
          if ((icnt-istart+1) .le. isize) then
            ii=0
            do ind=istart,icnt
              ii=ii+1
              i2a(ii)=iarr(ind)
            enddo
          else
            CALL wrerr(rtname,1,'Too many elements in param. '//param)
            istat=1
          endif
        CASE('I4','i4')
          icnt=isize
          istat = PILgetintvarvector(trim(param),icnt,iarr(1))
          if ((icnt-istart+1) .le. isize) then
            ii=0
            do ind=istart,icnt
              ii=ii+1
              i4a(ii)=iarr(ind)
            enddo
          else
            CALL wrerr(rtname,1,'Too many elements in param. '//param)
            istat=1
          endif
        CASE('R4','r4')
          icnt=isize
          istat = PILgetreal4varvector(trim(param),icnt,r4arr(1))
          if ((icnt-istart+1) .le. isize) then
            ii=0
            do ind=istart,icnt
              ii=ii+1
              r4a(ii)=r4arr(ind)
            enddo
          else
            CALL wrerr(rtname,1,'Too many elements in param. '//param)
            istat=1
          endif
        CASE('R8','r8')
          icnt=isize
          istat = PILgetrealvarvector(trim(param),icnt,r8arr(1))
          if ((icnt-istart+1) .le. isize) then
            ii=0
            do ind=istart,icnt
              ii=ii+1
              r8a(ii)=r8arr(ind)
            enddo
          else
            CALL wrerr(rtname,1,'Too many elements in param. '//param)
            istat=1
          endif
        CASE DEFAULT
          istat=1
          CALL wrerr(rtname,1,'Param. type '//cid//' not supported')
      END SELECT
          
    ENDIF
            
    IF (istat .ne. 0) THEN
      !           write(*,*) 'ISTAT:',istat
      CALL wrerr(rtname,1,'Error getting parameter: '//param)
      ierr=2
    ENDIF

    !      ENDIF
        
    !      istat = PILclose(istat)

    IF (istat .ne. 0) THEN
      ierr = 2
    ENDIF

  end subroutine rgetn_pil

  !---------------------------------------------------------------------------
  subroutine errstc(name, ipos)
    !---------------------------------------------------------------------------
    ! needed for compatibility with old SASS

    character(len=*) :: name
    integer :: ipos

  end subroutine errstc

  !----------------------------------------------------------------------------
  subroutine prep_gtcalf1(date_obs, date_end, strtdate, strtime, stpdate, stptime, chatter)
    !----------------------------------------------------------------------------
  
    character(len=19), intent(in) :: date_obs, date_end
    integer, intent(out) :: chatter
    character(len=10), intent(out) :: strtdate, strtime, stpdate, stptime
    character(len=6) :: cenv  

    ! stdheader times -> format expected by caldb
    strtdate = date_obs(1:10)
    stpdate = date_end(1:10)
    strtime = date_obs(12:19)
    stptime = date_end(12:19)

    ! SASS verbosity -> caldb verbosity
    call get_environment_variable("SASS_VERBOSITY", cenv)
    chatter = 0
    if (cenv == "sparse") chatter = 1
    if (cenv == "chatty") chatter = 5
    
  end subroutine prep_gtcalf1

  !----------------------------------------------------------------------------
  subroutine prep_gtcalf2(chatter)
    !----------------------------------------------------------------------------

    character(len=6) :: cenv
    integer, intent(out) :: chatter

    ! SASS verbosity -> caldb verbosity
    call get_environment_variable("SASS_VERBOSITY", cenv)
    chatter = 0
    if (cenv == "sparse") chatter = 1
    if (cenv == "chatty") chatter = 5

  end subroutine prep_gtcalf2

end module param
